"""Tests that the development and creation of fragments works as intended."""
import os
import unittest

import numpy as np
from setup_tests import setUpAll, tearDownAll

from pycoalescence.fragments import FragmentedLandscape, Fragment


def setUpModule():
    """
    Creates the output directory and moves logging files
    """
    setUpAll()


def tearDownModule():
    """
    Removes the output directory
    """
    tearDownAll()


class TestBasicFragmentedLandscape(unittest.TestCase):
    """Tests that a basic fragmented landscape works as intended."""

    def testPrecursoryValidation(self):
        """Tests that the precursory validation makes sense."""
        fl = FragmentedLandscape()
        with self.assertRaises(ValueError):
            fl._precursory_validation(number_fragments=100000, size=10, total=1)
        with self.assertRaises(ValueError):
            fl._precursory_validation(number_fragments=2, size=100, total=1)
        with self.assertRaises(ValueError):
            fl._precursory_validation(number_fragments=1, size=10, total=100000)

    def testAddRemaindersError(self):
        """Tests that adding the remainders works as intended."""
        fl = FragmentedLandscape()
        fl.total_added = 1
        fl.remainder = 1
        fl.number_fragments = 10
        with self.assertRaises(ValueError):
            fl._add_remainders()

    def testCheckPathAndFile(self):
        """Tests that an error is raised if the output file already exists."""
        fl = FragmentedLandscape()
        with self.assertRaises(IOError):
            fl._check_file_and_path()
        fl.output_file = os.path.join("sample", "SA_sample_fine.tif")
        with self.assertRaises(IOError):
            fl._check_file_and_path()

    def testDisabledSmoothing(self):
        """Tests that smoothing can be disabled correctly"""
        output = os.path.join("output", "frag_gen1.tif")
        fl = FragmentedLandscape(number_fragments=10, size=100, total=1000, output_file=output)
        fl.generate(override_smoothing=False)
        self.assertTrue(os.path.exists(output))

    def testEnabledSmoothing(self):
        """Tests that smoothing can be disabled correctly"""
        output = os.path.join("output", "frag_gen2.tif")
        fl = FragmentedLandscape(number_fragments=10, size=100, total=1000, output_file=output)
        fl.generate(override_smoothing=True)
        self.assertTrue(os.path.exists(output))

    def testFinalValidation(self):
        """Tests the errors raised during final validation."""
        fl = FragmentedLandscape()
        f1 = Fragment(1, 1)
        f2 = Fragment(-2, 2)
        f3 = Fragment(2, 2)
        fl.fragments = [f1, f2]
        fl.total_added = 1
        with self.assertRaises(ValueError):
            fl._final_validation()
        fl.total_added = 2
        with self.assertRaises(ValueError):
            fl._final_validation()
        fl.fragments = [f1, f3]
        fl.size = 3
        fl.remainder = 1
        with self.assertRaises(ValueError):
            fl._final_validation()
        fl.remainder = 0
        fl.grid = np.ones(shape=(2, 2))
        fl.total = 3
        with self.assertRaises(ValueError):
            fl._final_validation()



"""
Contains compatibility imports for Python 2.x.
"""

try:  # pragma: no cover
    FileExistsError = FileExistsError
    FileNotFoundError = FileNotFoundError
except NameError:  # pragma: no cover

    class FileExistsError(IOError):
        pass

    class FileNotFoundError(IOError):
        pass



"""
Simulate dispersal kernels on landscapes. Detailed :ref:`here <Simulate_landscapes>`.

:input:
    - Map file to simulate on
    - Set of dispersal pararameters, including the dispersal kernel, number of repetitions and landscape properties

:output:
    - Database containing each distance travelled so that metrics can be calculated.
    - A table is created for mean dispersal distance over a single step or for mean distance travelled.
"""
from __future__ import absolute_import

import logging
import os
import sys

from numpy import std
from numpy import zeros


try:
    try:
        import sqlite3
    except ImportError:  # pragma: no cover
        # Python 3 compatibility
        import sqlite as sqlite3
except ImportError as ie:  # pragma: no cover
    sqlite3 = None
    logging.warning("Problem importing sqlite module " + str(ie))

from pycoalescence.system_operations import check_parent
from pycoalescence.landscape import Landscape
from pycoalescence.necsim.necsim import CDispersalSimulator


class DispersalSimulation(Landscape):
    """
    Simulates a dispersal kernel upon a tif file to calculate landscape-level dispersal metrics.
    """

    def __init__(self, dispersal_db=None, file=None, logging_level=logging.WARNING):
        """
        Default initialiser for members of DispersalSimulation. Ensures that the database is

        :param str/pycoalescence.DispersalSimulation dispersal_db: path to a dispersal simulation database. Can also be
                                                                   a DispersalSimulation object containing the completed
                                                                   simulation.
        :param str file: sets the filename for reading tif files.
        :param bool is_sample: sets the sample mask to true, if it is a sampled file
        :param int logging_level: the level of logging to output during dispersal simulations
        """
        Landscape.__init__(self)
        self.logger = logging.Logger("necsim")
        self._create_logger(logging_level=logging_level)
        self._db_conn = None
        self.c_dispersal_simulation = None
        self._create_c_dispersal_simulation()
        self.deme = 1.0
        self.number_repeats = None
        self.number_steps = None
        self.seed = None
        self.number_workers = 1
        self.dispersal_method = None
        self.landscape_type = None
        self.sigma = None
        self.tau = None
        self.m_prob = None
        self.cutoff = None
        self.sequential = None
        self.dispersal_relative_cost = None
        self.restrict_self = None
        self.dispersal_file = None
        if file is not None:
            self.set_map(file)
        if isinstance(dispersal_db, DispersalSimulation):
            self.dispersal_database = dispersal_db.dispersal_database
        else:
            self.dispersal_database = dispersal_db

    def __del__(self):
        """
        Safely destroys the connection to the database, if it exists, and destroys the C++ objects.
        """
        self._close_database_connection()
        self.c_dispersal_simulation = None

    def _create_c_dispersal_simulation(self):
        """Creates the CDispersalSimulation object, if it has not already been created."""
        if self.c_dispersal_simulation is None:
            self.c_dispersal_simulation = CDispersalSimulator(self.logger)

    def _open_database_connection(self, database=None):
        """
        Opens the connection to the database, raising the appropriate errors if the database does not exist
        Should have a matching call to _close_database_connection() for safely destroying the connection to the database
        file.
        """
        if database is not None:
            if self._db_conn is not None:  # pragma: no cover
                return
            self.dispersal_database = database
        if self.dispersal_database is None:  # pragma: no cover
            raise ValueError("Dispersal database is not set, run test_average_dispersal() first or set dispersal_db.")
        if not os.path.exists(self.dispersal_database):
            raise IOError("Dispersal database does not exist: {}".format(self.dispersal_database))
        # Open the SQLite connection
        try:
            self._db_conn = sqlite3.connect(self.dispersal_database)
        except sqlite3.Error as e:  # pragma: no cover
            self._db_conn = None
            raise IOError("Error opening SQLite database: {}".format(e))

    def _close_database_connection(self):
        """Safely closes the database connection."""
        if self._db_conn is not None:
            try:
                self._db_conn.close()
                self._db_conn = None
            except sqlite3.Error as e:  # pragma: no cover
                self._db_conn = None
                raise IOError("Could not close database: " + str(e))

    def _check_table_exists(self, database=None, table_name="DISPERSAL_DISTANCES"):
        """
        Checks that the dispersal distances table exits, and returns true/false.

        :param database: the database to open

        :return: true if the DISPERSAL_DISTANCES table exists in the output database
        :rtype: bool
        """
        self._open_database_connection(database)
        existence = (
            self._db_conn.cursor()
            .execute("SELECT name FROM sqlite_master WHERE type='table' AND" " name='{}';".format(table_name))
            .fetchone()
            is not None
        )
        return existence

    def _check_output_database(self):
        """Sets the setup to false if the output database has not been generated already."""
        self._create_c_dispersal_simulation()
        if self.dispersal_database is None:
            raise ValueError("Dispersal database is None.")
        if not os.path.exists(self.dispersal_database):
            self.setup_complete = False
            self.c_dispersal_simulation.set_output_database(self.dispersal_database)

    def _remove_existing_db(self):
        """Removes an existing database correctly."""
        self._close_database_connection()
        self.c_dispersal_simulation = None
        self.setup_complete = False
        os.remove(self.dispersal_database)

    def set_map_files(
        self,
        fine_file,
        sample_file="null",
        coarse_file=None,
        historical_fine_file=None,
        historical_coarse_file=None,
        deme=1.0,
    ):
        """
        Sets the map files.

        Uses a null sampling regime, as the sample file should have no effect.

        :param str fine_file: the fine map file. Defaults to "null" if none provided
        :param str coarse_file: the coarse map file. Defaults to "none" if none provided
        :param str historical_fine_file: the historical fine map file. Defaults to "none" if none provided
        :param str historical_coarse_file: the historical coarse map file. Defaults to "none" if none provided
        :param int deme: the number of individuals per cell

        :rtype: None
        """
        Landscape.set_map_files(
            self,
            sample_file=sample_file,
            fine_file=fine_file,
            coarse_file=coarse_file,
            historical_fine_file=historical_fine_file,
            historical_coarse_file=historical_coarse_file,
        )
        self.deme = deme

    def set_dispersal_parameters(
        self,
        dispersal_method="normal",
        dispersal_file="none",
        sigma=1,
        tau=1,
        m_prob=1,
        cutoff=100,
        dispersal_relative_cost=1,
        restrict_self=False,
    ):
        """
        Sets the dispersal parameters.

        :param str dispersal_method: the dispersal method to use ("normal", "fat-tailed" or "norm-uniform")
        :param str dispersal_file: path to the dispersal map file, or none.
        :param float sigma: the sigma value to use for normal and norm-uniform dispersal
        :param float tau: the tau value to use for fat-tailed dispersal
        :param float m_prob: the m_prob to use for norm-uniform dispersal
        :param float cutoff: the cutoff value to use for norm-uniform dispersal
        :param float dispersal_relative_cost:relative dispersal ability through non-habitat
        :param bol restrict_self: if true, self-dispersal is prohibited
        """
        self.dispersal_method = dispersal_method
        self.sigma = sigma
        self.tau = tau
        self.m_prob = m_prob
        self.cutoff = cutoff
        self.dispersal_relative_cost = dispersal_relative_cost
        self.restrict_self = restrict_self
        self.dispersal_file = dispersal_file
        self.c_dispersal_simulation.set_dispersal_parameters(
            self.dispersal_method,
            self.dispersal_file,
            self.sigma,
            self.tau,
            self.m_prob,
            self.cutoff,
            self.dispersal_relative_cost,
            self.restrict_self,
        )

    def update_parameters(
        self,
        number_repeats=None,
        number_steps=None,
        seed=None,
        number_workers=None,
        dispersal_method=None,
        dispersal_file=None,
        sigma=None,
        tau=None,
        m_prob=None,
        cutoff=None,
        dispersal_relative_cost=None,
        restrict_self=None,
    ):
        """
        Provides a convenience function for updating all parameters which can be updated.

        :param int number_repeats: the number of repeats to perform the dispersal simulation for
        :param list/int number_steps: the number of steps to iterate for in calculating the mean distance travelled
        :param int seed: the random number seed
        :param int number_workers: the number of threads (>= 1) launched to run the distance simulation in parallel
        :param str dispersal_method: the method of dispersal
        :param str dispersal_file: the dispersal file (alternative to dispersal_method)
        :param float sigma: the sigma dispersal value
        :param float tau: the tau dispersal value
        :param float m_prob: the probability of drawing from a uniform distribution
        :param float cutoff: the maximum value for the uniform distribution
        :param float dispersal_relative_cost: the relative cost of moving through non-habitat
        :param bool restrict_self: if true, prohibits dispersal from the same cell

        :rtype: None
        """
        vars = locals().copy()
        for k, v in vars.items():
            if v is not None:
                setattr(self, k, v)
            if k == "number_steps" and v is not None:
                if isinstance(v, list):
                    self.number_steps = v
                else:
                    self.number_steps = [v]
        self.set_dispersal_parameters(
            self.dispersal_method,
            self.dispersal_file,
            self.sigma,
            self.tau,
            self.m_prob,
            self.cutoff,
            self.dispersal_relative_cost,
            self.restrict_self,
        )

    def set_simulation_parameters(
        self,
        number_repeats=None,
        output_database="output.db",
        seed=1,
        number_workers=1,
        dispersal_method="normal",
        landscape_type="closed",
        sigma=1,
        tau=1,
        m_prob=1,
        cutoff=100,
        sequential=False,
        dispersal_relative_cost=1,
        restrict_self=False,
        number_steps=1,
        dispersal_file="none",
    ):
        """
        Sets the simulation parameters for the dispersal simulations.

        :param int number_repeats: the number of times to iterate on the map
        :param str output_database: the path to the output database
        :param int seed: the random seed
        :param int number_workers: the number of threads (>= 1) launched to run the distance simulation in parallel
        :param str dispersal_method: the dispersal method to use ("normal", "fat-tailed" or "norm-uniform")
        :param str landscape_type: the landscape type to use ("infinite", "tiled_coarse", "tiled_fine", "clamped_coarse", "clamped_fine" or "closed")
        :param float sigma: the sigma value to use for normal and norm-uniform dispersal
        :param float tau: the tau value to use for fat-tailed dispersal
        :param float m_prob: the m_prob to use for norm-uniform dispersal
        :param float cutoff: the cutoff value to use for norm-uniform dispersal
        :param bool sequential: if true, end locations of one dispersal event are used as the start for the next. Otherwise,
        a new random cell is chosen
        :param float dispersal_relative_cost: relative dispersal ability through non-habitat
        :param bool restrict_self: if true, self-dispersal is prohibited
        :param list/int number_steps: the number to calculate for mean distance travelled, provided as an int or a list
                                       of ints
        :param str dispersal_file: path to the dispersal map file, or none.
        """
        self.number_repeats = number_repeats
        if output_database != "output.db" or self.dispersal_database is None:
            self.dispersal_database = output_database
        self.dispersal_database = os.path.abspath(self.dispersal_database)
        self.seed = seed
        self.number_workers = number_workers
        self.landscape_type = landscape_type
        self.sequential = sequential
        self.restrict_self = restrict_self
        if isinstance(number_steps, list):  # pragma: no cover
            self.number_steps = [int(x) for x in number_steps]
        else:
            self.number_steps = [int(number_steps)]
        if not os.path.exists(os.path.dirname(self.dispersal_database)):  # pragma: no cover
            os.makedirs(os.path.dirname(self.dispersal_database))
        self.c_dispersal_simulation.set_output_database(self.dispersal_database)
        self.set_dispersal_parameters(
            dispersal_method,
            dispersal_file,
            sigma,
            tau,
            m_prob,
            cutoff,
            dispersal_relative_cost,
            restrict_self,
        )

    def complete_setup(self):
        """
        Completes the setup for the dispersal simulation, including importing the map files and setting the historical
        maps.
        """
        if not self.is_setup_map:  # pragma: no cover
            raise RuntimeError("Maps have not been set up yet.")
        self._check_output_database()
        self.c_dispersal_simulation.set_dispersal_parameters(
            self.dispersal_method,
            self.dispersal_file,
            self.sigma,
            self.tau,
            self.m_prob,
            self.cutoff,
            self.dispersal_relative_cost,
            self.restrict_self,
        )
        if self.setup_complete:  # pragma: no cover
            self.logger.info("Set up has already been completed.")
        else:
            if len(self.historical_fine_list) != 0:
                self.c_dispersal_simulation.import_all_maps(
                    self.deme,
                    self.fine_map.file_name,
                    self.fine_map.x_size,
                    self.fine_map.y_size,
                    self.fine_map.x_offset,
                    self.fine_map.y_offset,
                    self.sample_map.x_size,
                    self.sample_map.y_size,
                    self.coarse_map.file_name,
                    self.coarse_map.x_size,
                    self.coarse_map.y_size,
                    self.coarse_map.x_offset,
                    self.coarse_map.y_offset,
                    int(self.coarse_scale),
                    self.landscape_type,
                    self.historical_fine_list,
                    [x for x in range(len(self.historical_fine_list))],
                    [float(x) for x in self.rates_list],
                    [float(x) for x in self.times_list],
                    self.historical_coarse_list,
                    [x for x in range(len(self.historical_fine_list))],
                    [float(x) for x in self.rates_list],
                    [float(x) for x in self.times_list],
                )
            else:
                self.c_dispersal_simulation.import_maps(
                    self.deme,
                    self.fine_map.file_name,
                    self.fine_map.x_size,
                    self.fine_map.y_size,
                    self.fine_map.x_offset,
                    self.fine_map.y_offset,
                    self.sample_map.x_size,
                    self.sample_map.y_size,
                    self.coarse_map.file_name,
                    self.coarse_map.x_size,
                    self.coarse_map.y_size,
                    self.coarse_map.x_offset,
                    self.coarse_map.y_offset,
                    int(self.coarse_scale),
                    self.landscape_type,
                )
            self.setup_complete = True

    def check_base_parameters(
        self,
        number_repeats=None,
        seed=None,
        sequential=None,
        number_workers=None,
        dispersal=False,
    ):
        """
        Checks that the parameters have been set properly.

        :param int number_repeats: the number of times to iterate on the map
        :param int seed: the random seed
        :param bool sequential: if true, runs repeats in the dispersal simulation sequentially
        :param int number_workers: the number of threads (>= 1) launched to run the distance simulation in parallel
        :prarm bool dispersal: True iff a dispersal instead of a distance simulation is to be run

        :rtype: None
        """
        if number_repeats is None and self.number_repeats is None:
            raise ValueError("number_repeats has not been set.")
        if seed is None and self.seed is None:
            raise ValueError("seed has not been set.")
        if sequential is None and self.sequential is None and dispersal:
            raise ValueError("sequential flag has not been set.")
        if number_workers is None and self.number_workers is None and not dispersal:
            raise ValueError("number_workers has not been set.")
        if number_repeats is not None:
            self.number_repeats = number_repeats
        if seed is not None:
            self.seed = seed
        if sequential is not None:
            self.sequential = sequential
        if number_workers is not None:
            self.number_workers = number_workers

        self._check_output_database()

    def run_mean_distance_travelled(self, number_repeats=None, number_steps=None, seed=None, number_workers=None):
        """
        Tests the dispersal kernel on the provided map, producing a database containing the average distance travelled
        after number_steps have been moved.

        .. note::

                mean distance travelled with number_steps=1 should be equivalent to running
                :func:`~run_mean_dispersal`

        :param int number_repeats: the number of times to iterate on the map
        :param int/list number_steps: the number of steps to take each time before recording the distance travelled
        :param int seed: the random seed
        :param int number_workers: the number of threads (>= 1) launched to run the distance simulation in parallel

        :rtype: None
        """
        self._close_database_connection()
        # Delete the file if it exists, and recursively create the folder if it doesn't
        check_parent(self.dispersal_database)
        if number_steps is None and self.number_steps in [None, [None], []]:
            raise ValueError("number_steps has not been set.")
        self.check_base_parameters(number_repeats=number_repeats, seed=seed, number_workers=number_workers)
        if number_steps is not None:
            self.number_steps = number_steps
        if not self.setup_complete:
            self.complete_setup()
        self.c_dispersal_simulation.run_mean_distance_travelled(
            self.number_repeats, self.number_steps, self.seed, self.number_workers
        )

    def run_all_distance_travelled(self, number_repeats=None, number_steps=None, seed=None, number_workers=None):
        """
        Tests the dispersal kernel on all cells on the provided map, producing a database containing the average distance
        travelled after number_steps have been moved.

        :param int number_repeats: the number of times to average over for each cell
        :param int/list number_steps: the number of steps to take each time before recording the distance travelled
        :param int seed: the random seed
        :param int number_workers: the number of threads (>= 1) launched to run the distance simulation in parallel

        :rtype: None
        """
        self._close_database_connection()
        # Delete the file if it exists, and recursively create the folder if it doesn't
        check_parent(self.dispersal_database)
        if number_steps is None and self.number_steps in [None, [None], []]:
            raise ValueError("number_steps has not been set.")
        self.check_base_parameters(number_repeats=number_repeats, seed=seed, number_workers=number_workers)
        if number_steps is not None:
            self.number_steps = number_steps
        if not self.setup_complete:
            self.complete_setup()
        self.c_dispersal_simulation.run_all_distance_travelled(
            self.number_repeats, self.number_steps, self.seed, self.number_workers
        )

    def run_sample_distance_travelled(
        self,
        samples_X,
        samples_Y,
        number_repeats=None,
        number_steps=None,
        seed=None,
        number_workers=None,
    ):
        """
        Tests the dispersal kernel on the sampled cells on the provided map, producing a database containing the average distance
        travelled after number_steps have been moved.

        :param list samples_X: list of the integer x coordinates of the sampled cells
        :param list samples_Y: list of the integer y coordinates of the sampled cells
        :param int number_repeats: the number of times to average over for each cell
        :param int/list number_steps: the number of steps to take each time before recording the distance travelled
        :param int seed: the random seed
        :param int number_workers: the number of threads (>= 1) launched to run the distance simulation in parallel

        :rtype: None
        """
        self._close_database_connection()
        # Delete the file if it exists, and recursively create the folder if it doesn't
        check_parent(self.dispersal_database)
        if number_steps is None and self.number_steps in [None, [None], []]:
            raise ValueError("number_steps has not been set.")
        self.check_base_parameters(number_repeats=number_repeats, seed=seed, number_workers=number_workers)
        if number_steps is not None:
            self.number_steps = number_steps
        if not self.setup_complete:
            self.complete_setup()
        self.c_dispersal_simulation.run_sample_distance_travelled(
            samples_X,
            samples_Y,
            self.number_repeats,
            self.number_steps,
            self.seed,
            self.number_workers,
        )

    def run_mean_dispersal(self, number_repeats=None, seed=None, sequential=None):
        """
        Tests the dispersal kernel on the provided map, producing a database containing each dispersal distance for
        analysis purposes.

        .. note:: should be equivalent to :func:`~run_mean_distance_travelled` with number_steps = 1

        :param int number_repeats: the number of times to iterate on the map
        :param int seed: the random seed
        :param bool sequential: if true, runs repeats sequentially
        """
        self._close_database_connection()
        # Delete the file if it exists, and recursively create the folder if it doesn't
        check_parent(self.dispersal_database)
        self.check_base_parameters(
            number_repeats=number_repeats,
            seed=seed,
            sequential=sequential,
            dispersal=True,
        )
        if not self.setup_complete:
            self.complete_setup()
        self.c_dispersal_simulation.run_mean_dispersal_distance(self.number_repeats, self.seed, self.sequential)

    def get_all_dispersal(self, database=None, parameter_reference=1):
        """
        Gets all mean dispersal values from the database if run_mean_dispersal has already been run.

        :raises: ValueError if dispersal_database is None and so run_mean_dispersal() has not been run
        :raises: IOError if the output database does not exist

        :param str database: the database to open
        :param int parameter_reference: the parameter reference to use (default 1)
        :return: the dispersal values from the database
        """
        if not self._check_table_exists(database=database, table_name="DISPERSAL_DISTANCES"):
            raise IOError("Database {} does not have a DISPERSAL_DISTANCES table".format(self.dispersal_database))
        try:
            self._open_database_connection(database=database)
            cursor = self._db_conn.cursor()
            sql_fetch = cursor.execute(
                "SELECT distance FROM DISPERSAL_DISTANCES WHERE parameter_reference = ?",
                (parameter_reference,),
            ).fetchall()
            if not sql_fetch:
                raise ValueError(
                    "Could not get dispersal distances for "
                    "parameter reference of {} from {}.".format(parameter_reference, self.dispersal_database)
                )
        except sqlite3.Error as e:  # pragma: no cover
            raise IOError("Could not get all dispersals from database: {}.".format(e))
        return [x[0] for x in sql_fetch]

    def get_mean_dispersal(self, database=None, parameter_reference=1):
        """
        Gets the mean dispersal for the map if run_mean_dispersal has already been run.

        :raises: ValueError if dispersal_database is None and so run_mean_dispersal() has not been run
        :raises: IOError if the output database does not exist

        :param str database: the database to open
        :param int parameter_reference: the parameter reference to use (default 1)).
        :return: mean dispersal from the database
        """
        if not self._check_table_exists(database=database, table_name="DISPERSAL_DISTANCES"):
            raise IOError("Database {} does not have a DISPERSAL_DISTANCES table".format(self.dispersal_database))
        try:
            self._open_database_connection(database=database)
            cursor = self._db_conn.cursor()
            sql_fetch = cursor.execute(
                "SELECT AVG(distance) FROM DISPERSAL_DISTANCES WHERE parameter_reference = ?",
                (parameter_reference,),
            ).fetchall()[0][0]
            if sql_fetch is None:
                raise ValueError(
                    "Could not get mean dispersal for "
                    "parameter reference of {} from {}.".format(parameter_reference, self.dispersal_database)
                )
        except sqlite3.Error as e:  # pragma: no cover
            raise IOError("Could not get mean dispersal from database: {}.".format(e))
        return sql_fetch

    def get_all_distances(self, database=None, parameter_reference=1):
        """
        Gets all total distances travelled from the database if run_mean_distance_travelled or
        run_all_distance_travelled or run_sample_distance_travelled has already been run.

        :raises: ValueError if dispersal_database is None and so run_mean_dispersal() has not been run
        :raises: IOError if the output database does not exist

        :param str database: the database to open
        :param int parameter_reference: the parameter reference to use (default 1)
        :return: the dispersal values from the database
        """
        if not self._check_table_exists(database=database, table_name="DISTANCES_TRAVELLED"):
            raise IOError("Database {} does not have a DISTANCES_TRAVELLED table".format(self.dispersal_database))
        try:
            self._open_database_connection(database=database)
            cursor = self._db_conn.cursor()
            sql_fetch = cursor.execute(
                "SELECT distance FROM DISTANCES_TRAVELLED WHERE parameter_reference = ?",
                (parameter_reference,),
            ).fetchall()
            if not sql_fetch:
                raise ValueError(
                    "Could not get distances travelled for "
                    "parameter reference of {} from {}.".format(parameter_reference, self.dispersal_database)
                )
        except sqlite3.Error as e:  # pragma: no cover
            raise IOError("Could not get all distances travelled from database: {}.".format(e))
        return [x[0] for x in sql_fetch]

    def get_distances_map(self, shape, database=None, parameter_reference=1):
        """
        Gets all total distances travelled from the database if run_mean_distance_travelled or
        run_all_distance_travelled or run_sample_distance_travelled has already been run
        and puts them inside a numpy matrix

        :raises: ValueError if dispersal_database is None and so run_mean_dispersal() has not been run
        :raises: IOError if the output database does not exist
        :raises: IndexError if the output database contains coordinates outside a matrix with shape=shape

        :param (int, int) shape: shape of the numpy matrix to return which will contain the distances
        :param str database: the database to open
        :param int parameter_reference: the parameter reference to use (default 1)
        :return: the dispersal values from the database
        """
        if not self._check_table_exists(database=database, table_name="DISTANCES_TRAVELLED"):
            raise IOError("Database {} does not have a DISTANCES_TRAVELLED table".format(self.dispersal_database))
        try:
            self._open_database_connection(database=database)
            cursor = self._db_conn.cursor()
            sql_fetch = cursor.execute(
                "SELECT x, y, distance FROM DISTANCES_TRAVELLED WHERE parameter_reference = ?",
                (parameter_reference,),
            ).fetchall()
            if not sql_fetch:
                raise ValueError(
                    "Could not get distances travelled for "
                    "parameter reference of {} from {}.".format(parameter_reference, self.dispersal_database)
                )
        except sqlite3.Error as e:  # pragma: no cover
            raise IOError("Could not get all distances travelled from database: {}.".format(e))

        dmap = zeros(shape=shape)

        for x, y, dist in sql_fetch:
            dmap[y, x] = dist

        return dmap

    def get_mean_distance_travelled(self, database=None, parameter_reference=1):
        """
        Gets the mean dispersal for the map if run_mean_distance_travelled or run_all_distance_travelled or
        run_sample_distance_travelled has already been run.

        :raises: ValueError if dispersal_database is None and so test_average_dispersal() has not been run
        :raises: IOError if the output database does not exist

        :param str database: the database to open
        :param int parameter_reference: the parameter reference to use (or 1 for default parameter reference).
        :return: mean of dispersal from the database
        """
        if not self._check_table_exists(database=database, table_name="DISTANCES_TRAVELLED"):
            raise IOError("Database {} does not have a DISTANCES_TRAVELLED table".format(self.dispersal_database))
        try:
            self._open_database_connection(database=database)
            cursor = self._db_conn.cursor()
            sql_fetch = cursor.execute(
                "SELECT AVG(distance) FROM DISTANCES_TRAVELLED WHERE parameter_reference = ?",
                (parameter_reference,),
            ).fetchall()[0][0]
            if sql_fetch is None:
                raise ValueError(
                    "Could not get mean distance travelled for "
                    "parameter reference of {} from {}.".format(parameter_reference, self.dispersal_database)
                )
        except sqlite3.Error as e:  # pragma: no cover
            raise IOError("Could not get average distance from database: {}.".format(e))
        return sql_fetch

    def get_stdev_dispersal(self, database=None, parameter_reference=1):
        """
        Gets the standard deviation of dispersal for the map if run_mean_dispersal has already been run.

        :raises: ValueError if dispersal_database is None and so test_average_dispersal() has not been run
        :raises: IOError if the output database does not exist

        :param str database: the database to open
        :param int parameter_reference: the parameter reference to use (or 1 for default parameter reference).
        :return: standard deviation of dispersal from the database
        """
        if not self._check_table_exists(database=database, table_name="DISPERSAL_DISTANCES"):
            raise IOError("Database {} does not have a DISPERSAL_DISTANCES table".format(self.dispersal_database))
        try:
            self._open_database_connection(database=database)
            cursor = self._db_conn.cursor()
            sql_fetch = [
                x[0]
                for x in cursor.execute(
                    "SELECT distance FROM DISPERSAL_DISTANCES " "WHERE parameter_reference = ?",
                    (parameter_reference,),
                ).fetchall()
            ]
            if len(sql_fetch) == 0:  # pragma: no cover
                raise ValueError("No distances in DISPERSAL_DISTANCES, cannot find standard deviation.")
            stdev_distance = std(sql_fetch)
        except sqlite3.Error as e:  # pragma: no cover
            raise IOError("Could not get average distance from database: {}".format(e))
        return stdev_distance

    def get_stdev_distance_travelled(self, database=None, parameter_reference=1):
        """
        Gets the standard deviation of the distance travelled for the map if run_mean_distance_travelled or
        run_all_distance_travelled or run_sample_distance_travelled has already
        been run.

        :raises: ValueError if dispersal_database is None and so test_average_dispersal() has not been run
        :raises: IOError if the output database does not exist

        :param str database: the database to open
        :param int parameter_reference: the parameter reference to use (or 1 for default parameter reference).

        :return: standard deviation of dispersal from the database
        :rtype: float
        """
        if not self._check_table_exists(database=database, table_name="DISTANCES_TRAVELLED"):
            raise IOError("Database {} does not have a DISTANCES_TRAVELLED table".format(self.dispersal_database))
        try:
            self._open_database_connection(database=database)
            cursor = self._db_conn.cursor()
            sql_fetch = [
                x[0]
                for x in cursor.execute(
                    "SELECT distance FROM DISTANCES_TRAVELLED" " WHERE parameter_reference = ?",
                    (parameter_reference,),
                ).fetchall()
            ]
            if len(sql_fetch) == 0:  # pragma: no cover
                raise ValueError("No distances in DISTANCES_TRAVELLED, cannot find standard deviation.")
            stdev_distance = std(sql_fetch)
        except sqlite3.Error as e:  # pragma: no cover
            raise IOError("Could not get average distance from database: {}.".format(e))
        return stdev_distance

    def get_database_parameters(self, reference=None):
        """
        Gets the dispersal simulation parameters from the dispersal_db

        :param reference: the reference to obtain parameters for

        :return: the dispersal simulation parameters

        :rtype: dict
        """
        self._open_database_connection()
        try:
            cursor = self._db_conn.cursor()
            cursor.execute(
                "SELECT ref, simulation_type, sigma, tau, m_prob, cutoff, dispersal_method, map_file, seed,"
                " number_steps, number_repeats FROM PARAMETERS"
            )
        except sqlite3.Error as e:  # pragma: no cover
            raise IOError("Could not get dispersal simulation parameters from database: {}".format(e))
        column_names = [member[0] for member in cursor.description]
        main_dict = {}
        for row in cursor.fetchall():
            values = [x for x in row]
            # Python 2.x support
            if sys.version_info[0] != 3:  # pragma: no cover
                for i, each in enumerate(values):
                    if isinstance(each, unicode):
                        values[i] = each.encode("ascii")
            # Now convert it into a dictionary
            main_dict[values[0]] = dict(zip(column_names[1:], values[1:]))
        if reference is None:
            return main_dict
        else:
            try:
                return main_dict[reference]
            except KeyError:
                raise KeyError("No reference exists in the database with a value of {}".format(reference))

    def get_database_references(self):
        """
        Gets the references from the database.

        :return: a list of references from the database

        :rtype: list
        """
        self._open_database_connection()
        try:
            cursor = self._db_conn.cursor()
            cursor.execute("SELECT DISTINCT(ref) FROM PARAMETERS")
        except sqlite3.Error as e:  # pragma: no cover
            raise IOError("Could not get dispersal simulation parameters from database: {}".format(e))
        return [x[0] for x in cursor.fetchall()]



"""Tests spatial_algorithms module for spatial routines."""
import unittest

from pycoalescence.spatial_algorithms import *

try:
    import osgeo

    gdal_version = osgeo.__version__
except ImportError as ie:
    gdal_version = "0.0.0"
    logging.warn("Could not import osgeo: {}".format(ie))


class TestSpatialAlgorthims(unittest.TestCase):
    """Tests all the spatial algorithms"""

    def testCalculateDistanceBetween(self):
        """Tests that the distance between points is correctly calculated"""
        self.assertAlmostEqual(calculate_distance_between(0, 0, 1, 1), 2 ** 0.5, places=8)
        self.assertAlmostEqual(calculate_distance_between(1, 10, 8, 10), 7, places=8)
        self.assertAlmostEqual(calculate_distance_between(1, -1, 10, -1), 9, places=8)
        self.assertAlmostEqual(calculate_distance_between(-1, -1, 1, 1), 8 ** 0.5, places=8)

    def testCalculateCentreMass(self):
        """Tests that the centre of mass is correctly calculated for a list of points."""
        points_list = [[-1, -1], [1, -1], [1, 1], [-1, 1]]
        self.assertTupleEqual((0, 0), calculate_centre_of_mass(points_list))
        points_list = [[0, -1], [-1, 0], [0, 1], [1, 0]]
        self.assertTupleEqual((0, 0), calculate_centre_of_mass(points_list))
        with self.assertRaises(ValueError):
            points_list = [[-1, -1], [1, -1], [-1, 1], [1, 1]]
            self.assertListEqual((0, 0), calculate_centre_of_mass(points_list))

    def testReflectDimensions(self):
        """Tests that dimensions are correctly reflected"""
        points = reflect_dimensions([[1, 1], [0.5, 0.5]], maximums=(2, 2))
        expected_points = [[1, 1], [0.5, 0.5]]
        expected_points.extend([[1, -1], [0.5, -0.5], [1, 3], [0.5, 3.5], [-1, 1], [-0.5, 0.5], [3, 1], [3.5, 0.5]])
        for i, each in enumerate(expected_points):
            self.assertAlmostEqual(each[0], points[i][0], places=3)
            self.assertAlmostEqual(each[1], points[i][1], places=3)

    def testArchimedesSpiral(self):
        """Tests a single iteration of the archimedes spiral produces the expected output"""
        self.assertEqual(archimedes_spiral(0, 0, 1, 2 * 3.14)[0], 0)
        self.assertEqual(archimedes_spiral(0, 0, 1, 2 * 3.14)[1], -1)

    def testConvertCoordinates(self):
        """Tests that conversion of coordinates works as expected between two different coordinate systems."""
        input_srs = osr.SpatialReference()
        input_srs.ImportFromWkt(
            'PROJCS["WGS_1984_UTM_Zone_50N",GEOGCS["GCS_WGS_1984",DATUM["WGS_1984",'
            'SPHEROID["WGS_84",6378137,298.257223563]],'
            'PRIMEM["Greenwich",0],'
            'UNIT["Degree",0.017453292519943295],'
            'AUTHORITY["EPSG","4326"]],'
            'PROJECTION["Transverse_Mercator"],'
            'PARAMETER["latitude_of_origin",0],'
            'PARAMETER["central_meridian",117],'
            'PARAMETER["scale_factor",0.9996],'
            'PARAMETER["false_easting",500000],'
            'PARAMETER["false_northing",0],'
            'UNIT["Meter",1],'
            'AUTHORITY["EPSG","32650"]]'
        )
        if int(gdal_version[0]) < 3:

            output_srs = osr.SpatialReference(
                wkt='GEOGCS["WGS 84",DATUM["WGS_1984",'
                'SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],'
                'AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],'
                'UNIT["degree",0.0174532925199433],AUTHORITY["EPSG","4326"]]'
            )
            x, y = convert_coordinates(600544.9967818621, 644607.5784286809, input_srs, output_srs)
        else:
            # TODO check if this is a bug with gdal 3.0 or gdal 2.0
            output_srs = osr.SpatialReference(
                'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],'
                'AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433],AUTHORITY["EPSG","4326"]]'
            )
            y, x = convert_coordinates(600544.9967818621, 644607.5784286809, input_srs, output_srs)
        self.assertAlmostEqual(117.9082031776, x)
        self.assertAlmostEqual(5.8310327261971, y)

    def testEstimationSigmaFromDistance(self):
        """Tests the estimation of sigma from a distance travelled and a number of steps."""
        self.assertAlmostEqual(1.0, estimate_sigma_from_distance(1.0 * (math.pi / 2.0) ** 0.5, 1), places=7)
        self.assertAlmostEqual(2.0, estimate_sigma_from_distance(2.0 * (math.pi / 2.0) ** 0.5, 1), places=7)
        self.assertAlmostEqual(4.0, estimate_sigma_from_distance(4.0 * (math.pi / 2.0) ** 0.5, 1), places=7)
        self.assertAlmostEqual(1.0, estimate_sigma_from_distance((2 * math.pi / 2.0) ** 0.5, 2), places=7)
        self.assertAlmostEqual(2.0, estimate_sigma_from_distance(2 * (2 * math.pi / 2.0) ** 0.5, 2), places=7)
        self.assertAlmostEqual(4.0, estimate_sigma_from_distance(4 * (2 * math.pi / 2.0) ** 0.5, 2), places=7)
        self.assertAlmostEqual(2.0, estimate_sigma_from_distance(2 * (100 * math.pi / 2.0) ** 0.5, 100), places=7)
        self.assertAlmostEqual(4.0, estimate_sigma_from_distance(4 * (100 * math.pi / 2.0) ** 0.5, 100), places=7)



"""
Compile **necsim** with default or provided compilation options. Intended for internal usage during ``pip`` or ``conda``
builds, although manual installation is also possible by running this file from the command line.
``python installer.py`` configures the install by detecting system components and compiles the ``C++`` files, if
possible. Command line flags can be provided to installer.py to modify the install (see
:ref:`Compilation Options <sec Compilation Options>` for more information).
"""

from __future__ import print_function, absolute_import, division  # Only Python 2.x

import itertools
import logging
import os
import platform
import re
import shutil
import subprocess
import sys
import time


# Import system operations for subprocess executation and log file handling

try:
    from .future_except import FileNotFoundError
    from .system_operations import execute_log_info, set_logging_method, execute_silent, mod_directory
except (ImportError, SystemError, ValueError):  # pragma: no cover
    from future_except import FileNotFoundError
    from system_operations import execute_log_info, set_logging_method, execute_silent, mod_directory
from setuptools.command.build_ext import build_ext
from distutils import sysconfig


def get_lib_and_gdal():
    gdal_inc_path = None
    gdal_dir = None
    if platform.system() == "Windows":
        libdir = get_python_library("{}.{}".format(sys.version_info.major, sys.version_info.minor))
    else:
        conda_prefix = os.environ.get("PREFIX", None)  # for conda only under unix
        libdir = sysconfig.get_config_var("LIBDIR")
        if conda_prefix is not None:
            gdal_inc_path = os.path.join(conda_prefix, "include")
            gdal_dir = os.path.join(conda_prefix, "lib")
        else:
            try:
                gdal_inc_path = subprocess.check_output(["gdal-config", "--cflags"], env=os.environ)
                gdal_dir = subprocess.check_output(["gdal-config", "--prefix"], env=os.environ)
            except subprocess.CalledProcessError:
                pass
            if gdal_inc_path is not None:
                gdal_inc_path = gdal_inc_path.decode("utf-8").split(" ")[0][2:].replace("\n", "")
            if gdal_dir is not None:
                gdal_dir = gdal_dir.decode("utf-8").replace("\n", "")
    return libdir, gdal_inc_path, gdal_dir


class Installer(build_ext):  # pragma: no cover
    """Wraps configuration and compilation of C++ code."""

    def __init__(self, dist, **kwargs):
        """Generates the link to the mod directory for installation."""
        build_ext.__init__(self, dist)
        self.mod_dir = mod_directory
        self.build_dir = None
        self.obj_dir = None

    def make_depend(self):
        """
        Runs make depend in the lib directory to calculate all dependencies for the header and source files.

        .. note:: Fails silently if makedepend is not installed, printing an error to logging.
        """
        try:
            execute_silent(["make", "depend"], cwd=os.path.join(self.mod_dir, "lib/"))
        except (RuntimeError, subprocess.CalledProcessError) as rte:
            logging.error("Could not execute makedepend: " + str(rte))
            logging.error("Using default dependencies instead.")
            self.use_default_depends()

    def create_default_depend(self):
        """
        Runs the default makedepend command, outputting dependencies to lib/depends_default.

        Used to generate a default dependency file on a system where makedepend exists, for a system where it does not.
        """
        try:
            execute_silent(
                ["makedepend", "*.cpp", "necsim/*.cpp", "-f" "depends_default", "-p", "obj/", "-Y"],
                cwd=os.path.join(self.mod_dir, "lib/"),
            )
        except (RuntimeError, subprocess.CalledProcessError) as rte:
            logging.error("Could not execute makedepend: " + str(rte))

    def use_default_depends(self):
        """
        Uses the default dependencies, copying all contents of depends_default to the end of Makefile.

        .. note:: Zero error-checking is done here as the Makefiles should not change, and the depends_default file should
                  be created using create_default_depend()

        """
        # First read Makefile
        with open(os.path.join(self.mod_dir, "lib", "Makefile"), "r") as f_in:
            lines = f_in.readlines()
        with open(os.path.join(self.mod_dir, "lib", "depends_default"), "r") as f_default:
            lines_default = f_default.readlines()
        # Now write all back out
        with open(os.path.join(self.mod_dir, "lib/Makefile"), "w") as f_out:
            for line in lines:
                if "# DO NOT DELETE" in line:
                    break
                f_out.write(line)
            for line in lines_default:
                f_out.write(line)

    def do_compile(self):
        """
        Compiles the C++ necsim program by running make. This changes the working directory to wherever the module has been
        installed for the subprocess call.
        """
        # Check that the make file exists
        if os.path.exists(os.path.join(self.mod_dir, "lib/")):
            time.sleep(0.5)
            self.make_depend()
            # Sleep to ensure that file timings are updated (support for HPC systems with inaccurate timings).
            time.sleep(1)
            try:
                execute_log_info(["make", "all"], cwd=os.path.join(self.mod_dir, "lib/"))
                logging.info("Compilation exited successfully.")
            except (RuntimeError, subprocess.CalledProcessError) as rte:
                logging.error(str(rte))
                logging.error("Compilation attempted, but error thrown.")
        else:
            raise IOError("C++ library does not exist! Check relative file path")

    def move_shared_object_file(self):
        """
        Moves the shared object (.so) file to the build directory.
        :return:
        """
        directory = os.path.join(self.mod_dir, "necsim")
        dirv = "sharedpy" + sys.version[0]
        for file in ["libnecsim.so"]:
            src = os.path.join(self.mod_dir, "lib", file)
            version_dir = os.path.join(directory, dirv)
            if not os.path.exists(src):
                raise IOError("Shared object file {} does not exists. Check installation was successful.".format(src))
            if not os.path.exists(directory):
                os.mkdir(directory)
            if not os.path.exists(version_dir):
                os.mkdir(version_dir)
            if not os.path.exists(os.path.join(version_dir, "__init__.py")):
                open(os.path.join(version_dir, "__init__.py"), "a").close()
            dst = os.path.join(directory, dirv, file)
            if os.path.exists(dst):
                os.remove(dst)
            shutil.copy(src, dst)

    def get_build_dir(self):
        """
        Gets the build directory.

        :return: the build directory path
        """
        if self.build_dir is None:
            directory = os.path.join(self.mod_dir, "necsim")
            return directory
        return os.path.join(self.build_dir)

    def get_obj_dir(self):
        """
        Gets the obj directory for installing obj files to.

        :return: the obj directory path
        """
        if self.obj_dir is None:
            return "obj"
        return self.obj_dir

    def configure(self, opts=None):
        """
        Runs ./configure --opts with the supplied options. This should create the makefile for compilation, otherwise a
        RuntimeError will be thrown.

        :param opts: a list of options to pass to the ./configure call
        """
        if "--with-debug" in opts or self.debug:
            for i, each in enumerate(opts):
                if "-DNDEBUG" in each:
                    opts[i] = opts[i].replace("-DNDEBUG", "")
        if os.path.exists(os.path.join(self.mod_dir, "lib", "configure")):
            try:
                if opts is None:
                    execute_log_info(
                        [
                            "./configure",
                            "--with-verbose",
                            "BUILDDIR={}".format(self.get_build_dir()),
                            "OBJDIR={}".format(self.get_obj_dir()),
                        ],
                        cwd=os.path.join(self.mod_dir, "lib"),
                    )
                else:
                    command = ["./configure"]
                    command.extend(opts)
                    if "BUILDDIR" not in opts:
                        command.append("BUILDDIR={}".format(self.get_build_dir()))
                    if "OBJDIR" not in opts:
                        command.append("OBJDIR={}".format(self.get_obj_dir()))
                    execute_log_info(command, cwd=os.path.join(self.mod_dir, "lib"))
            except subprocess.CalledProcessError as cpe:
                raise subprocess.CalledProcessError("Configuration attempted, but error thrown: ".format(cpe))
        else:
            raise RuntimeError("File src/configure does not exist. Check installation has been successful.")

    def autoconf(self):
        """
        Runs the `autoconf` bash function (assuming that autoconf is available) to create the `configure` executable.
        """
        try:
            execute_log_info(["autoconf"], cwd=os.path.join(self.mod_dir, "lib/"))
        except (RuntimeError, subprocess.CalledProcessError, FileNotFoundError, OSError, IOError) as cpe:
            logging.warning(
                "Could not run autoconf function to generate configure executable. "
                "Please run this functionality manually if installation fails."
            )
            logging.warning(str(cpe))

    def clean(self):
        """
        Runs make clean in the NECSim directory to wipe any previous potential compile attempts.
        """
        try:
            time.sleep(0.5)
            execute_log_info(["make", "obj_dir"], cwd=os.path.join(self.mod_dir, "lib/"))
            execute_log_info(["make", "build_dir"], cwd=os.path.join(self.mod_dir, "lib/"))
            execute_log_info(["make", "clean"], cwd=os.path.join(self.mod_dir, "lib/"))
        except subprocess.CalledProcessError as cpe:
            raise RuntimeError("Make file has not been generated. Cannot clean: ".format(cpe))

    def get_ldflags(self):
        """Get the ldflags that Python was compiled with, removing some problematic options."""
        sysldflags = sysconfig.get_config_var("LDFLAGS")
        if sysldflags is None:
            ldflags = ""
        else:
            ldflags = re.sub(r"-arch \b[^ ]*[\ ]*", "", sysldflags) + " "
        if "--sysroot=" in ldflags:
            logging.warning("--sysroot found in LDFLAGS, removing")
            ldflags = re.sub(r"--sysroot=.*[,\s]", "", ldflags)
        ldflags = ldflags.replace("\n", " ")
        return ldflags

    def get_ldshared(self):
        """Get the ldshared Python variables and replaces -bundle with -shared for proper compilation."""
        ldflags = sysconfig.get_config_var("LDSHARED")
        if ldflags is None:
            return ""
        ldflags = " ".join(ldflags.split()[1:]).replace("-bundle", "-shared")
        return ldflags

    def get_compilation_flags(self, display_warnings=False):
        """
        Generates the compilation flags for passing to ./configure.
        :param display_warnings: If true, runs with the -Wall flag for compilation (displaying all warnings). Default is False.

        :return: list of compilation flags.
        :rtype: list
        """
        # Get the relevant flags that Python was originally compiled with, to be passed to the C++ code.
        include = str("CPPFLAGS=-I" + sysconfig.get_python_inc()).replace("\n", "")
        cflags = " " + sysconfig.get_config_var("CFLAGS")
        cflags = str(re.sub(r"-arch \b[^ ]*", "", cflags)).replace("\n", "")  # remove any architecture flags
        cflags += " "
        py_ldflags = str(
            "-L"
            + sysconfig.get_python_lib(standard_lib=True)
            + " -L"
            + sysconfig.get_config_var("DESTDIRS").replace(" ", " -L")
        ).replace("\n", "")
        py_lib = "PYTHON_LIB=-lpython"
        ldflags = str("LDFLAGS=" + self.get_ldflags())
        # Get the shared object platform-specific compilation flags.
        platform_so = "PLATFORM_SO="
        if platform.system() == "Linux":
            platform_so += "-shared"
        elif platform.system() == "Darwin":
            platform_so += "-dynamiclib"
        elif platform.system() == "Windows":
            raise SystemError(
                "COMPILATION FAILURE: Windows is not yet supported. You must compile the libraries yourself."
            )
        else:
            logging.critical("OS not detected, compilation failures likely. Please report this error.")
        # Make sure that the linker directs to the correct Python library (such as -lpython3.5m)
        # Eventually this will also detect if the install is for an anaconda distribution.
        if "conda" not in sys.version and "Continuum" not in sys.version:
            py_lib += sys.version[0:3]
            if "m" in sysconfig.get_config_var("LIBRARY"):
                py_lib += "m"
        else:
            cflags += "-I{}/include {}".format(os.environ["PREFIX"], "--enable-shared")
            ldflags += "-L{}/lib".format(os.environ["PREFIX"])
            py_lib += sys.version[0:3]
            if "m" in sysconfig.get_config_var("LIBRARY"):
                py_lib += "m"
            cflags += " -DANACONDA"
        ldflags += " " + py_ldflags
        py_ldflags = "PYTHON_LDFLAGS=" + py_ldflags
        call = [include + cflags, py_lib, ldflags, py_ldflags, platform_so]
        # Remove the flags which would potentially cause unnecessary warnings to be thrown.
        # This can be disabled by passing display_warnings=True
        if not display_warnings:
            call = [re.sub("-Wstrict-prototypes|-Wno-unused-result|-Wunused-variable|-Wall", "", x) for x in call]
        return call

    def run_configure(self, argv=None, logging_level=logging.INFO, display_warnings=False):
        """
        Configures the install for compile options provided via the command line, or with default options if no options exist.
        Running with ``-help`` or `-h` will display the compilation configurations called from ``./configure``.

        :param argv: the arguments to pass to configure script
        :param logging_level: the logging level to utilise (defaults to INFO).
        :param display_warnings: If true, runs with the -Wall flag for compilation (displaying all warnings). Default is False.
        """
        if argv is None:
            argv = [None]
        call = self.get_compilation_flags(display_warnings=display_warnings)
        set_logging_method(logging_level=logging_level)
        self.autoconf()
        if len(argv) == 1:
            logging.info("No compile options provided, using defaults.")
            print("Obj: {}".format(self.get_obj_dir()))
            print("Build: {}".format(self.get_build_dir()))
            call.extend(["--with-verbose", "OBJDIR={}", "BUILDDIR={}".format(self.get_obj_dir(), self.get_build_dir())])
        else:
            if argv[1] == "--h" or argv[1] == "-h" or argv[1] == "-help" or argv[1] == "--help":
                execute_log_info(["./configure", "--help"], cwd=os.path.join(self.mod_dir, "lib/"))
            if isinstance(argv, str):
                call.append(argv[2])
            else:
                call.extend(argv[1:])
        logging.info(call)
        self.configure(opts=call)
        self.clean()

    def backup_makefile(self):
        """
        Copies the makefile to a saved folder so that even if the original is overwritten, the last successful
        compilation can be recorded.
        """
        src = os.path.join(self.mod_dir, "lib", "Makefile")
        dst = os.path.join(self.mod_dir, "reference", "Makefile")
        if not os.path.exists(src):
            logging.error("Makefile does not exist at {}. Check successful compilation".format(src))
            return
        if os.path.exists(dst):
            os.remove(dst)
        shutil.copy2(src, dst)

    def copy_makefile(self):
        """
        Copies the backup makefile to the main directory, if it exists.
        Throws an IOError if no makefile is found.
        """
        src = os.path.join("reference", "Makefile")
        dst = os.path.join("lib", "Makefile")
        if not os.path.exists(src):
            raise IOError("Cannot find backup Makefile, requires running of configuration scripts.")
        if os.path.exists(dst):
            os.remove(dst)
        shutil.copy2(src, dst)

    def configure_and_compile(self, argv=[None], logging_level=logging.INFO):
        """
        Calls the configure script, then runs the compilation.

        :param argv: the arguments to pass to configure script
        :param logging_level: the logging level to utilise (defaults to INFO).
        :rtype: None
        """
        set_logging_method(logging_level=logging_level)
        display_warnings = "display_warnings=True" in argv
        self.run_configure(argv, logging_level=logging_level, display_warnings=display_warnings)
        self.do_compile()
        self.backup_makefile()

    def setuptools_cmake(self, ext):
        """
        Configures cmake for setuptools usage.

        :param ext: the extension to build cmake on
        """
        if "conda" not in sys.version and "Continuum" not in sys.version:
            extdir = os.path.abspath(
                os.path.join(os.path.dirname(self.get_ext_fullpath(ext.name)), "pycoalescence", "necsim")
            )
        else:
            sp_dir = os.environ.get("SP_DIR")
            if sp_dir is None:
                sp_dir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
            extdir = os.path.join(sp_dir, "pycoalescence", "necsim")
        cmake_args, build_args = self.get_default_cmake_args(extdir)
        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get("CXXFLAGS", ""), self.distribution.get_version())
        if "INTEL_LICENSE_FILE" in env.keys():
            env["CXX"] = "icpc"
            env["CC"] = "icc"
            cmake_args.extend(["-DCMAKE_C_COMPILER=icc", "-DCMAKE_CXX_COMPILER=icpc", "-DUSING_INTEL=ON"])
        self.run_cmake(ext.sourcedir, cmake_args, build_args, self.build_temp, env)

    def run_cmake(self, src_dir, cmake_args, build_args, tmp_dir, env):
        """
        Runs cmake to compile necsim.

        :param src_dir: the source directory for necsim .cpp and .h files
        :param cmake_args: arguments to pass to the cmake project
        :param tmp_dir: the build directory to output cmake files to
        :param env: the os.environ (or other environmental variables) to pass on
        """
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)
        try:
            subprocess.check_call(["cmake", src_dir] + cmake_args, cwd=tmp_dir, env=env)
            if platform.system() == "Windows":
                subprocess.check_call(["nmake"], cwd=tmp_dir, env=env)
                # subprocess.check_call(["nmake", "install"], cwd=tmp_dir, env=env)
            else:
                subprocess.check_call(
                    ["cmake", "--build", ".", "--target", "necsim"] + build_args, cwd=tmp_dir, env=env
                )
        except subprocess.CalledProcessError as cpe:
            raise SystemError("Fatal error running cmake in directory: {}".format(cpe))
        if platform.system() == "Windows":
            shutil.copy(os.path.join(tmp_dir, "necsim.pxd"), os.path.join(self.get_build_dir(), "libnecsim.pyd"))

    def run(self):
        """Runs installation and generates the shared object files - entry point for setuptools"""
        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        """Builds the C++ and Python extension."""
        self.build_dir = os.path.abspath(
            os.path.join(os.path.dirname(self.get_ext_fullpath(ext.name)), "pycoalescence", "necsim")
        )
        self.obj_dir = self.build_dir
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        self.setuptools_cmake(ext)

    def clean_cmake(self):
        """Deletes the cmake files and object locations if they exist."""
        for path in [
            os.path.join(self.get_build_dir(), "libnecsim.so"),
            os.path.join(self.get_build_dir(), "libnecsim.so.dSYM"),
            os.path.join(self.get_build_dir(), "libnecsim.pyd"),
            os.path.join(self.get_build_dir(), "libnecsim.dylib"),
            os.path.join(self.mod_dir, "obj"),
        ]:
            if os.path.exists(path):
                if os.path.isdir(path):
                    shutil.rmtree(path)
                else:
                    os.remove(path)

    def get_default_cmake_args(self, output_dir):
        """
        Returns the default cmake configure and build arguments.

        :param output_dir: the output directory to use

        :return: tuple of two lists, first containing cmake configure arguments, second containing build arguments
        :rtype: tuple
        """
        cfg = "Debug" if self.debug else "Release"
        cflags = sysconfig.get_config_var("CFLAGS")
        if cflags is not None:
            cflags = str(re.sub(r"-arch \b[^ ]*", "", cflags)).replace("\n", "")  # remove any architecture flags
        else:
            cflags = ""
        libdir, gdal_inc_path, gdal_dir = get_lib_and_gdal()
        if libdir is None:
            libdir = os.path.abspath(os.path.join(sysconfig.get_config_var("LIBDEST"), "..", "libs"))
            if sysconfig.get_config_var("LIBDEST") == None:
                raise SystemError("Cannot detect library directory for Python install.")
        cmake_args = [
            "-DPYTHON_LIBRARY:FILEPATH={}".format(libdir),
            "-DPYTHON_CPPFLAGS:='{}'".format(cflags),
            "-DPYTHON_LDFLAGS:='{}'".format(self.get_ldshared()),
            "-DPYTHON_INCLUDE_DIR:FILEPATH={}".format(sysconfig.get_python_inc()),
            "-DCMAKE_BUILD_TYPE={}".format(cfg),
        ]
        if gdal_inc_path is not None:
            cmake_args.append("-DGDAL_INCLUDE_DIR={}".format(gdal_inc_path))
        if gdal_dir is not None:
            cmake_args.append("-DGDAL_DIR={}".format(gdal_dir))
        build_args = ["--config", cfg]
        if platform.system() == "Windows":
            cmake_args += [
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY:PATH={}".format(output_dir),
                "-DCMAKE_PREFIX_PATH:PATH={}".format(os.environ.get("LIBRARY_PREFIX", "")),
                "-DCMAKE_INSTALL_PREFIX:PATH={}".format(os.environ.get("LIBRARY_PREFIX", "")),
                "-DCMAKE_BUILD_TYPE:STRING={}".format(cfg),
                "-G",
                "NMake Makefiles",
            ]
            # if sys.maxsize > 2 ** 32: # Removed to allow for building on conda-forge
            #     cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:

            cmake_args += [
                "-DCMAKE_BUILD_TYPE={}".format(cfg),
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY:PATH={}".format(output_dir),
            ]
            build_args += ["--", "-j2"]
        return cmake_args, build_args


def get_python_library(python_version):  # pragma: no cover
    """Get path to the Python library associated with the current Python
    interpreter."""
    # determine direct path to libpython
    python_library = sysconfig.get_config_var("LIBRARY")
    potential_library = None
    # if static (or nonexistent), try to find a suitable dynamic libpython
    if python_library is None or python_library[-2:] == ".a":

        candidate_lib_prefixes = ["", "lib"]

        candidate_extensions = [".lib", ".so"]
        if sysconfig.get_config_var("WITH_DYLD"):
            candidate_extensions.insert(0, ".dylib")

        candidate_versions = [python_version]
        if python_version:
            candidate_versions.append("")
            candidate_versions.insert(0, "".join(python_version.split(".")[:2]))

        abiflags = getattr(sys, "abiflags", "")
        candidate_abiflags = [abiflags]
        if abiflags:
            candidate_abiflags.append("")

        # Ensure the value injected by virtualenv is
        # returned on windows.
        # Because calling `sysconfig.get_config_var('multiarchsubdir')`
        # returns an empty string on Linux, `du_sysconfig` is only used to
        # get the value of `LIBDIR`.
        libdir = sysconfig.get_config_var("LIBDIR")
        if sysconfig.get_config_var("MULTIARCH"):
            masd = sysconfig.get_config_var("multiarchsubdir")
            if masd:
                if masd.startswith(os.sep):
                    masd = masd[len(os.sep) :]
                libdir = os.path.join(libdir, masd)

        if libdir is None:
            libdir = os.path.abspath(os.path.join(sysconfig.get_config_var("LIBDEST"), "..", "libs"))

        candidates = (
            os.path.join(libdir, "".join((pre, "python", ver, abi, ext)))
            for (pre, ext, ver, abi) in itertools.product(
                candidate_lib_prefixes, candidate_extensions, candidate_versions, candidate_abiflags
            )
        )
        for candidate in candidates:
            if os.path.exists(candidate):
                # we found a (likely alternate) libpython
                potential_library = candidate
                if potential_library[-2:] != ".a":
                    break
    # Otherwise still a static library, keep searching
    if potential_library is None:
        raise IOError("No Python libraries found")
    return potential_library


if __name__ == "__main__":  # pragma: no cover
    fail = True
    from distutils.dist import Distribution
    import argparse

    parser = argparse.ArgumentParser(description="Build the C++ library (necsim) required for pycoalescence.")
    parser.add_argument("--cmake", action="store_true", default=True, dest="cmake", help="use the cmake build process")
    parser.add_argument(
        "--autotools",
        action="store_true",
        default=False,
        dest="autotools",
        help="Use the autotools build process (./configure and make)",
    )
    parser.add_argument(
        "--compiler-args",
        metavar="N",
        type=str,
        nargs="+",
        dest="compiler_args",
        default=[],
        help="Additional arguments to pass to the autotools compiler",
    )
    parser.add_argument(
        "--cmake-args",
        metavar="N",
        type=str,
        nargs="+",
        dest="cmake_args",
        default=[],
        help="Additional arguments to pass to the cmake compiler during configuration",
    )
    parser.add_argument(
        "--cmake-build-args",
        metavar="N",
        type=str,
        nargs="+",
        dest="cmake_build_args",
        default=[],
        help="Additional arguments to pass to the cmake compiler at build time",
    )
    parser.add_argument("--debug", action="store_true", default=False, dest="debug", help="Compile using DEBUG defines")
    parser.add_argument(
        "-c",
        "-C",
        "--compile",
        action="store_true",
        default=False,
        dest="compile_only",
        help="Compile only, do not re-configure necsim",
    )
    parser.add_argument(
        "--clean",
        action="store_true",
        default=False,
        dest="clean",
        help="Clean previous cmake builds from this directory.",
    )

    args = parser.parse_args()
    if args.cmake and args.autotools:
        raise ValueError("Cannot use both cmake and autotools build process - specify one or the other.")
    if not args.cmake or args.autotools:
        raise ValueError("Must specify compilation either using autotools or cmake.")
    dist = Distribution()
    installer = Installer(dist)
    installer.debug = args.debug
    if args.clean:
        installer.clean_cmake()
    else:
        if args.cmake:
            env = os.environ.copy()
            build_dir = installer.get_build_dir()
            obj_dir = installer.get_obj_dir()
            cmake_args, build_args = installer.get_default_cmake_args(build_dir)
            cmake_args += args.cmake_args
            build_args += args.cmake_build_args
            src_dir = os.path.join(installer.mod_dir, "lib")
            installer.run_cmake(src_dir, cmake_args, build_args, obj_dir, env)
        else:
            if platform.system() != "Windows":
                raise SystemError("Usage of configure and make on a windows system is not supported.")
            if args.compile_only:
                set_logging_method(logging_level=logging.INFO)
                installer.copy_makefile()
                installer.do_compile()
                fail = False
            if fail:
                installer.configure_and_compile(args.compiler_args)



"""
Port older simulation outputs to the updated naming conventions. Should not be required by most users.
"""

import sqlite3

from pycoalescence.sqlite_connection import check_sql_table_exist


def update_parameter_names(database):
    """
    Alters the parameters names of SIMULATION_PARAMETERS in the database so that it matches the
    updated naming convention.

    Provided for back-compatibility with older simulations.

    .. note:: If the simulation does not require updating, this function exits silently.

    :param database: the database path to alter the names of

    :return: None

    :rtype: None
    """
    try:
        db = sqlite3.connect(database)
        c = db.cursor()
        if not check_sql_table_exist(db, "SIMULATION_PARAMETERS"):
            raise IOError("Table SIMULATION_PARAMETERS does not exist in database.")
        # Check if the parameters are already updated
        sql_query = (
            "SELECT seed, task, output_dir, speciation_rate, sigma, tau, deme,"
            " sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate,"
            " gen_since_historical, time_config_file, coarse_map_file, coarse_map_x, coarse_map_y,"
            " coarse_map_x_offset, coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, "
            "fine_map_y, fine_map_x_offset, fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y,"
            "sample_x_offset, sample_y_offset, historical_coarse_map, historical_fine_map, sim_complete, "
            "dispersal_method, m_probability, cutoff, landscape_type, protracted, "
            "min_speciation_gen, max_speciation_gen, dispersal_map FROM SIMULATION_PARAMETERS"
        )
        try:
            c.execute(sql_query)
            db = None
            return
        except sqlite3.Error:
            pass
        sql_query = "ALTER TABLE SIMULATION_PARAMETERS RENAME TO SIM_P_backup; "
        try:
            c.execute(sql_query)
        except sqlite3.Error:  # pragma: no cover
            c.execute("DROP TABLE SIM_P_backup;")
            c.execute(sql_query)
        sql_query = "CREATE TABLE SIMULATION_PARAMETERS(seed INT PRIMARY KEY not null, task INT NOT NULL,"
        sql_query += (
            "output_dir TEXT NOT NULL, speciation_rate DOUBLE NOT NULL, sigma DOUBLE NOT NULL,tau DOUBLE NOT NULL,"
            " deme INT NOT NULL, "
        )
        sql_query += (
            "sample_size DOUBLE NOT NULL, max_time INT NOT NULL, dispersal_relative_cost DOUBLE NOT NULL, "
            "min_num_species "
        )
        sql_query += "INT NOT NULL, habitat_change_rate DOUBLE NOT NULL, gen_since_historical DOUBLE NOT NULL, "
        sql_query += (
            "time_config_file TEXT NOT NULL, coarse_map_file TEXT NOT NULL, coarse_map_x INT NOT NULL, "
            "coarse_map_y INT NOT NULL,"
        )
        sql_query += (
            "coarse_map_x_offset INT NOT NULL, coarse_map_y_offset INT NOT NULL, coarse_map_scale DOUBLE NOT "
            "NULL, fine_map_file TEXT NOT NULL, fine_map_x INT NOT NULL,"
        )
        sql_query += "fine_map_y INT NOT NULL, fine_map_x_offset INT NOT NULL, fine_map_y_offset INT NOT NULL, "
        sql_query += "sample_file TEXT NOT NULL, grid_x INT NOT NULL, grid_y INT NOT NULL, sample_x INT NOT NULL, "
        sql_query += "sample_y INT NOT NULL, sample_x_offset INT NOT NULL, sample_y_offset INT NOT NULL, "
        sql_query += (
            "historical_coarse_map TEXT NOT NULL, historical_fine_map TEXT NOT NULL, sim_complete INT NOT NULL, "
        )
        sql_query += "dispersal_method TEXT NOT NULL, m_probability DOUBLE NOT NULL, cutoff DOUBLE NOT NULL, "
        sql_query += "restrict_self INT NOT NULL, landscape_type TEXT NOT NULL, protracted INT NOT NULL, "
        sql_query += (
            "min_speciation_gen DOUBLE NOT NULL, max_speciation_gen DOUBLE NOT NULL, dispersal_map TEXT NOT NULL);"
        )
        c.execute(sql_query)
        try:
            sql_query = (
                "INSERT INTO SIMULATION_PARAMETERS(seed, task, output_dir, speciation_rate, sigma, tau, deme,"
                " sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate, gen_since_historical,"
                " time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, "
                "coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset,"
                "fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, "
                "historical_coarse_map, historical_fine_map, sim_complete, dispersal_method, m_probability, cutoff, "
                "restrict_self, landscape_type, protracted, min_speciation_gen, max_speciation_gen, dispersal_map)"
                " SELECT seed, job_type, output_dir, speciation_rate, sigma, tau, deme,"
                " sample_size, max_time, dispersal_relative_cost, min_num_species, forest_change_rate, time_since_pristine,"
                " time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, "
                "coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset,"
                "fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, "
                "pristine_coarse_map, pristine_fine_map, sim_complete, dispersal_method, m_probability, cutoff, "
                "restrict_self, infinite_landscape, protracted, min_speciation_gen, max_speciation_gen, dispersal_map FROM SIM_P_backup;"
            )
            c.execute(sql_query)
        except sqlite3.Error:
            # Provide additional support for a different naming convention
            try:
                sql_query = (
                    "INSERT INTO SIMULATION_PARAMETERS(seed, task, output_dir, speciation_rate, sigma, tau, deme,"
                    " sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate, gen_since_historical,"
                    " time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, "
                    "coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset,"
                    "fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, "
                    "historical_coarse_map, historical_fine_map, sim_complete, dispersal_method, m_probability, cutoff, "
                    "restrict_self, landscape_type, protracted, min_speciation_gen, max_speciation_gen, dispersal_map)"
                    " SELECT seed, job_type, output_dir, speciation_rate, sigma, tau, deme,"
                    " sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate, time_since_pristine,"
                    " time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, "
                    "coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset,"
                    "fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, "
                    "pristine_coarse_map, pristine_fine_map, sim_complete, dispersal_method, m_probability, cutoff, "
                    "restrict_self, infinite_landscape, protracted, min_speciation_gen, max_speciation_gen, dispersal_map FROM SIM_P_backup;"
                )
                c.execute(sql_query)
            except sqlite3.Error:
                # Provide additional support for a different naming convention
                sql_query = (
                    "INSERT INTO SIMULATION_PARAMETERS(seed, task, output_dir, speciation_rate, sigma, tau, deme,"
                    " sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate, gen_since_historical,"
                    " time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, "
                    "coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset,"
                    "fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, "
                    "historical_coarse_map, historical_fine_map, sim_complete, dispersal_method, m_probability, cutoff, "
                    "restrict_self, landscape_type, protracted, min_speciation_gen, max_speciation_gen, dispersal_map)"
                    " SELECT seed, job_type, output_dir, speciation_rate, sigma, tau, deme,"
                    " sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate, gen_since_pristine,"
                    " time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, "
                    "coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset,"
                    "fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, "
                    "pristine_coarse_map, pristine_fine_map, sim_complete, dispersal_method, m_probability, cutoff, "
                    "restrict_self, infinite_landscape, protracted, min_speciation_gen, max_speciation_gen, dispersal_map FROM SIM_P_backup;"
                )
                c.execute(sql_query)
        c.execute("DROP TABLE SIM_P_backup;")
        db.commit()
        db.close()
        db = None
    except sqlite3.Error as soe:  # pragma: no cover
        raise sqlite3.Error("Could not update simulation parameters: {}".format(soe))
